""" Exceptions generated by either PXTL implementation.
"""

import sys, string, types, traceback
from pxtl.constants import *

class PXTLError(Exception):
  """ Base class for all errors caused by bad code in templates. Parameter
      'context' is a DOM Node indicating the bit of the template that caused the
      error.
  """
  def __init__(self, context):
    Exception.__init__(self)
    self.context= context
    self.location= None
    if hasattr(context, 'location'):
      self.location= context.location
  def __str__(self):
    try:
      desc= '%s, at: %s' % (self.describeError(), self.describeContext())
      if type(desc)==UnicodeType:
        return string.join(map(lambda c: (c, '?')[ord(c)>=128], desc), '')
      return desc
    except:
      traceback.print_exc()
      raise
  def describeError(self):
    return 'PXTL error'
  def describeContext(self):
    if self.context==None:
      return '(unknown)'
    node= self.context
    if node.nodeType==node.ELEMENT_NODE:
      desc= '<%s>' % node.tagName
      node= node.parentNode
    elif node.nodeType==node.ATTRIBUTE_NODE:
      desc= 'Attribute "%s"' % node.name
      node= node.ownerElement
    elif node.nodeType==node.PROCESSING_INSTRUCTION_NODE:
      desc= 'PI "%s"' % node.target
      node= node.parentNode
    else:
      desc= node.nodeName
    while node.nodeType==node.ELEMENT_NODE:
      desc= '%s in <%s>' % (desc, node.tagName)
      node= node.parentNode
    try:
      location= self.context.location
      if location.uri is not None:
        desc= desc+', '+location.uri
      if location.lineNumber!=-1:
        desc= desc+', line '+str(location.lineNumber)
      if location.columnNumber!=-1:
        desc= desc+' char '+str(location.columnNumber)
    except AttributeError:
      pass
    return desc


# Generic exceptions
#
class PXTLSyntaxError(PXTLError):
  """ Template is not valid PXTL.
  """
  def describeError(self):
    return "Invalid PXTL syntax"

class PXTLNameError(PXTLError):
  """ Template includes item with unknown name.
  """
  def __init__(self, context, name):
    PXTLError.__init__(self, context)
    self.name= name
  def describeError(self):
    return "unknown name %s" % self.name

class PXTLTypeError(PXTLError):
  """ A PXTL attribute evaluated to the wrong type. Parameter 'given' is what
      was evaluated, 'needed' is a type object specifying what was expected.
  """
  def __init__(self, context, given, needed):
    PXTLError.__init__(self, context)
    self.given= given
    self.needed= needed
  def describeError(self):
    return '%s required, got %s (%s)' % map(str,
      (self.needed, type(self.given), self.given)
    )

class PXTLValueError(PXTLError):
  """ A PXTL attribute evaluated to an unknown value. Parameter 'given' is what
      was evaluated, 'needed' is a list of possible values.
  """
  def __init__(self, context, given, needed):
    PXTLError.__init__(self, context)
    self.given= given
    self.needed= needed
  def describeError(self):
    return '%s or %s required, got %s' % (
      string.join(map(str, self.needed[:-1]), ', '),
      str(self.needed[-1]), str(self.given)
    )

class PXTLRuntimeError(PXTLError):
  """ An error occurred whilst executing Python code from a template. 'trace'
      is the traceback of the root-cause error.
  """
  def __init__(self, context, exc_info, code):
    PXTLError.__init__(self, context)
    self.error= exc_info[1]
    self.trace= exc_info[2]
    self.code= code
  def describeError(self):
    return '%s: %s' % (str(self.error.__class__), self.error)

# Specific expections
#
class PXTLParseError(PXTLError):
  """ A template couldn't be parsed by the DOM implementation. If it is pxdom,
      we can read the location of the parse error. 
  """
  def __init__(self, error, documentURI):
    context= None
    if hasattr(error, 'location'):
      context= NullNode()
      context.location= error.location
      context.ownerDocument= NullNode()
      context.ownerDocument.documentURI= documentURI
    PXTLError.__init__(self, context)
    self.error= error
    self.documentURI= documentURI
  def describeError(self):
    return '%s: %s' % (str(self.error.__class__), str(self.error))
  def describeContext(self):
    return 'document '+self.documentURI

class PXTLDocumentTypeError(PXTLTypeError):
  def __init__(self, context, given):
    PXTLTypeError.__init__(self, context, given, types.TupleType)

class PXTLMethodValueError(PXTLValueError):
  def __init__(self, given):
    PXTLValueError.__init__(self, None, given, serialisers.keys())

class PXTLTagNameError(PXTLNameError):
  def __init__(self, context):
    PXTLNameError.__init__(self, context, context.tagName)

class PXTLAttributeNameError(PXTLNameError):
  def __init__(self, context):
    PXTLNameError.__init__(self, context, context.name)

class PXTLPITargetNameError(PXTLNameError):
  def __init__(self, context):
    PXTLNameError.__init__(self, context, context.target)

class PXTLNoParserError(PXTLSyntaxError):
  def describeError(self):
    return 'a DOM L3 Core/Load implementation is needed for import elements'

class PXTLAttributeMissingError(PXTLSyntaxError):
  def __init__(self, context, name):
    PXTLSyntaxError.__init__(self, context)
    self.name= name
  def describeError(self):
    return '%s required' % self.name

class PXTLMultipleConditionalError(PXTLSyntaxError):
  def describeError(self):
    return 'each element may only have one PXTL conditional attribute'

class PXTLTagnameAttrError(PXTLTypeError):
  def __init__(self, attr, given):
    PXTLTypeError.__init__(self, attr, given, types.StringType)

class PXTLTagnamespaceAttrError(PXTLValueError):
  def __init__(self, attr, given):
    PXTLValueError.__init__(self, attr, given, ['anything but PXTL or xmlns'])

class PXTLForItemsError(PXTLSyntaxError):
  def describeError(self):
    return "for element must have exactly one 'in' or 'range' attribute"

class PXTLForRangeError(PXTLTypeError):
  def __init__(self, attr, given):
    PXTLTypeError.__init__(self, attr, given, types.TupleType)

class PXTLWhileMinError(PXTLTypeError):
  def __init__(self, attr, given):
    PXTLTypeError.__init__(self, attr, given, types.IntType)

class PXTLPseudoPIOpenError(PXTLSyntaxError):
  def __init__(self, attr, ix):
    PXTLSyntaxError.__init__(self, attr)
    self.ix= ix
  def describeError(self):
    return 'pseudo-PI left open, beginning %s...' % self.context.value[self.ix:self.ix+9]

class PXTLPseudoPITargetNameError(PXTLNameError):
  def __init__(self, context, name):
    PXTLNameError.__init__(self, context, name)
  def describeError(self):
    return 'unknown PXTL pseudo-PI target %s' % self.name

class PXTLPseudoPITargetPXError(PXTLSyntaxError):
  def describeError(self):
    return 'only PXTL PIs can be used as pseudo-PIs, not %s' % self.name

class PXTLAttrTypeError(PXTLTypeError):
  def __init__(self, attr, given):
    PXTLTypeError.__init__(self, attr, given, types.DictType)

class PXTLAttrNameTypeError(PXTLTypeError):
  def __init__(self, attr, given):
    PXTLTypeError.__init__(self, attr, given, types.StringType)

class PXTLFutureFeatureError(PXTLSyntaxError):
  def __init__(self, context, feature):
    PXTLSyntaxError.__init__(self, context)
    self.feature= feature
  def describeError(self):
    return 'feature %s unimplemented in this version of Python' % self.feature

class PXTLCallTypeError(PXTLTypeError):
  def __init__(self, context, given):
    PXTLTypeError.__init__(self, context, given, types.InstanceType)
  def describeError(self):
    return 'call fn must be an object defined by the def element, not %s'

class PXTLCallArgsError(PXTLValueError):
  def __init__(self, context, given):
    PXTLValueError.__init__(self, context, given, types.NoneType)
  def describeError(self):
    if self.given==None:
      return 'subtemplate function requires arguments'
    else:
      return 'subtemplate function does not take arguments'

class PXTLConditionalOrderError(PXTLSyntaxError):
  def __init__(self, context, old_cond):
    PXTLSyntaxError.__init__(self, context)
    self.old_cond= old_cond
  def describeError(self):
    if self.old_cond==None:
      return '%s must follow a conditional clause' % self.context.localName
    return '%s cannot follow %s' % (self.context.localName, self.old_cond)

class PXTLImportGlobalsError(PXTLTypeError):
  def __init__(self, context, given):
    PXTLTypeError.__init__(self, context, given, types.DictType)

class NullNode:
  pass
